MODULE myPML
use nrtype
implicit  none

real, allocatable :: eta(:,:), phi(:,:)

contains
! eta has padding width of npml+1 around the central domain:
!   ? [pml] [main domain] [pml] ?
! phi has padding width of npml:
!     [pml] [main domain] [pml]
! if free surface,  upper edge has no pml padding; eta has 1 pixel width,
!     etaz(1) got to be 0 because it's to be max'ed with etax
!----------------------------
subroutine  extend_domain_2_pml(Aou, Ain, npml, nx, nz, FREESURF)
! A's:   fast dim. x
! Ain:   nx x nz
! Aou:  (nx+2npml) x (nz+{1,2}npml)
real,      intent(out)::  Aou(:,:)    ! vdt2 encompasses pml region
real,      intent(in) ::  Ain(:,:)
integer,   intent(in) ::  npml, nx, nz
logical,   intent(in) ::  FREESURF
integer    :: ix, iz, z2

z2 = npml
if (FREESURF)   z2 = 0

Aou(npml+1 : npml+nx, z2+1 : z2+nz) = Ain

! next, extend:
! left & right edges
!DEC$ IVDEP
do iz = z2+1, z2+nz
  Aou(1:npml, iz) = Aou(npml+1, iz)
  Aou(npml+nx+1 : nx+2*npml, iz) = Aou(npml+nx, iz)
end do
! extend bottom edge
!DEC$ IVDEP
do iz = z2+nz+1, z2+nz+npml
  Aou(:, iz) = Aou(:, z2+nz)
end do
! extend up
if (.not. FREESURF) then
!DEC$ IVDEP
  do iz = 1, npml
    Aou(:, iz) = Aou(:, npml+1)
  end do
end if
end subroutine
!-------------
subroutine  init_pml(velmax, dx, dt, nx, nz, npml, FREESURF, memsz)
 real,      intent(in) ::  velmax, dx, dt
 integer,   intent(in) ::  nx, nz, npml
 logical,   intent(in) ::  FREESURF
 real,      intent(out)::  memsz
 
 real,      allocatable::  etax(:), etaz(:)
 
 allocate( etax( -npml: nx+npml+1))
 
 if (FREESURF) then
   allocate(eta( -npml: nx+npml+1,     0: nz+npml+1),  etaz(    0: nz+npml+1))
   allocate(phi(1-npml: nx+npml,       1: nz+npml))
 else
   allocate(eta( -npml: nx+npml+1, -npml: nz+npml+1),  etaz(-npml: nz+npml+1))
   allocate(phi(1-npml: nx+npml,  1-npml: nz+npml))
 endif
   memsz = size(eta,1)*size(eta,2)*4. + &
	       size(phi,1)*size(phi,2)*4.
 call pml_profile(etax, etaz, velmax, dx, npml, FREESURF)
 call pml_outermax_profile(eta, etax, etaz, dt)
 deallocate(etax, etaz)
end subroutine
!-------------
subroutine kill_pml
   deallocate(eta, phi)
end subroutine
!-------------
subroutine pml_outermax_profile(eta, etax, etaz, dt)
real,  intent(out) :: eta(:,:)
real,  intent(in)  :: etax(:), etaz(:), dt

integer            :: i, j, lenx, lenz
lenx= size(etax)
lenz= size(etaz)
do j= 1, lenz
  do i= 1, lenx
    eta(i,j) = max(etax(i), etaz(j)) * dt  ! eta absorbs dt, for convenience in PDE
  enddo                                    ! eta is now dimensionless.
enddo
end subroutine

!----------------------------
subroutine pml_profile(etax, etaz, velmax, dx, npml, FREESURF)
real,  intent(out)   ::  etax(:), etaz(:)
real,  intent(in)    ::  velmax,  dx
integer,  intent(in) ::  npml
logical,   intent(in)::  FREESURF

integer    :: i, lenx, lenz
real       :: ftmp, etabc

etax = 0.
etaz = 0.
lenx = size(etax)
lenz = size(etaz)
etabc= 3.*velmax*log(1000.)/(2.*dx*real(npml))
do i = 1 , npml+1
  ftmp = (real(i)/real(npml+1))**2 * etabc
  etax(lenx - npml +i-1) = ftmp
  etaz(lenz - npml +i-1) = ftmp
  etax(npml+2-i)         = ftmp
  if (.not. FREESURF)    etaz(npml+2-i) = ftmp
enddo
end subroutine
!----------------------------
END MODULE

