!-----------------------
!  version:     1.0
!  author:      Yunsong Huang
!  email:       yunsong.huang@kaust.edu.sa
!  updated:     01-29-2013
!  purpose:     2d FDTD with PML
!-----------------------


MODULE   fdtd_2d

use datatype0;  use nrtype;       use myMPI0
use myPML;      use source_multi; use seism_uti_module
implicit none
PRIVATE
real            :: coef(0:4)      ! FD stencil, 1-side.  has 1/dx^2
real            :: lap, hdx_2, dx_2
real, target,   allocatable :: wvext1(:,:), wvext2(:,:), &
                  wvfld1(:,:), wvfld2(:,:), Laplac(:,:)
! when a pointer is assigned to a whole array variable, it inherits the 
! lower bounds of the variable, otherwise, the lower bounds default to 1.
! http://en.wikipedia.org/wiki/Fortran_95_language_features#Pointers_as_dynamic_aliases
!
! So, don't want to declare a higher dim. array and assign array section to a
! pointer, which would lost the array's lower bound.
! Instead, declare individual array sections to begin with, and make pointers
! point to them, thus inheriting the bounds.
real, pointer   :: p2(:,:), p1(:,:), pb(:,:), pa(:, :),  ptmp(:,:)
real,           allocatable ::       vdt2(:,:)
      ! wvext     two pressure wavefields, for fdtd wave propagation, w/ pml & lx
	  ! wvfld     ditto, w/o pml.  For source wavefield reconstruction.  As
	  !           boundaries are saved, no worries about pml layers.
	  ! vdt2      (v dt)^2,         w/  pml, required by all fdtd routines!
		 
type  fdtd_bound
 real,allocatable::u(:,:,:), d(:,:,:), l(:,:,:), r(:,:,:) 
      !  up & dn:  nx+2 X  lx       X nt-2
      !  lf & rt:  lx X (nz+2-2*lx) X nt-2
	  ! Recall: 
	  !  No need to reconstruct PML region, only inner region is of interest.
	  !  As wavefield of last two time steps are saved, no need to save boundary
      !	 (thus 'nt-2').   
	  !  x is fast dim., which is preferred to extend, so up & dn extend across
	  !  the whole x-domain (thus 'nx').  In contrast, lf & rt don't extend over
	  !  the whole z-domain, because the corners have been covered by up & dn.
	  !
	  !  Update: to compute spatial Laplacian for inner domain, we need width-1
	  !  layer in addition.  So, nx+2 in stead of nx,  nz+2 instead of nz.
end type
type(fdtd_bound):: Bnd
integer         :: k, i               ! temp loop variables
integer         :: x1,x2,x3,x4,x5,x6, z1,z2,z3,z4,z5,z6, lx, &
                   xb1,xb2,xb3,xb4,xb5,xb6, zb1,zb2,zb3,zb4,zb5,zb6 
				   ! 'b': associated with back reconstruction of source wavefield
				   ! [1,2],  [3,4],  [5,6]  define the domain's regions
#ifdef DEBUGfwi				   
  character(len=100) :: tmpfnm
#endif
public          :: init_fdtd, kill_fdtd, update_vdt2, forward_a_bundle, &
				   INIT_GRADIENT, gradient_a_bundle
contains 
!===================================
function init_fdtd(md)   result(mem_size)
real                         :: mem_size
type(domain_param),intent(in):: md

x1   = md%x1;   x2   = md%x2;   x3   = md%x3
x4   = md%x4;   x5   = md%x5;   x6   = md%x6
z1   = md%z1;   z2   = md%z2;   z3   = md%z3
z4   = md%z4;   z5   = md%z5;   z6   = md%z6 
lx   = md%lx

xb1  = x3 - 1;  xb3  = xb1+lx;  xb2  = xb3-1
xb6  = x4 + 1;  xb4  = xb6-lx;  xb5  = xb4+1

zb1  = z3 - 1;  zb3  = zb1+lx;  zb2  = zb3-1
zb6  = z4 + 1;  zb4  = zb6-lx;  zb5  = zb4+1

coef     =  0.
if   (lx == 2)  then
  coef(0)= -5.;        coef(1) = 4./3.;    coef(2) = -1./12. 
else
  coef(0)= -205./36.;  coef(1) = 8./5.;    coef(2) = -1./5. 
                       coef(3) = 8./315.;  coef(4) = -1./560. 
endif
coef     =  coef/ (md%dx * md%dx)
hdx_2    =  1./(4.*md%dx * md%dx)
dx_2     =  1./   (md%dx * md%dx)
allocate(Bnd%u(xb1    :xb6,    zb1    :zb2,    1:md%nt-2),&
	     Bnd%d(xb1    :xb6,    zb5    :zb6,    1:md%nt-2),&
	     Bnd%l(xb1    :xb2,    zb3    :zb4,    1:md%nt-2),&
	     Bnd%r(xb5    :xb6,    zb3    :zb4,    1:md%nt-2),&
		 wvext1(x1-lx :x6+lx,  z1-lx  :z6+lx),            &
		 wvext2(x1-lx :x6+lx,  z1-lx  :z6+lx),            &
		 vdt2(  x1    :x6,     z1     :z6   ),            &
		 wvfld1(xb1   :xb6,    zb1    :zb6  ),            &
		 wvfld2(xb1   :xb6,    zb1    :zb6  ),            &
		 Laplac(x3    :x4,     z3     :z4   ))
		   
mem_size = (2.*size(Bnd%u) + 2.*size(Bnd%l) + size(vdt2)+ &
            2.*size(wvext1)+ 2.*size(wvfld1)+ size(Laplac))*4. ! bytes
end function
!-------------
subroutine  kill_fdtd()
 deallocate(Bnd%u, Bnd%d, Bnd%l, Bnd%r,  wvext1, wvext2, vdt2, &
            wvfld1, wvfld2, Laplac)
end subroutine
!-------------
subroutine  update_vdt2(v, dt, npml, nx, nz, FREESURF)
 ! needs to be called everytime v is updated, 
 ! including the very 1st time v is initialized.
 real,      intent(in) ::  v(:,:),   dt  ! v is not extended
 integer,   intent(in) ::  npml, nx, nz
 integer(1),intent(in) ::  FREESURF

 call extend_domain_2_pml(vdt2, v, npml, nx, nz, FREESURF==1)
 vdt2 = vdt2 * dt
 vdt2 = vdt2 * vdt2
#ifdef DEBUGwv
   call write_binfile('log/vdt2.bin', vdt2)
#endif 
end subroutine
!-------------
subroutine put_free_surface(u, lx1, x0, x7, z0, z7) 
! Enforce free surfaces.  When there's PML, this is not important, but nice to have
! a correct way of handling the boundaries.  When PML is absent, this is required.
integer,   intent(in)    :: lx1, x0, x7, z0, z7 ! lx1: FDTD spatial order/2 - 1,
! e.g., lx1 = 4-1 = 3, when FDTD ~ O(2,8)
! x0, x7,  z0, z7:  where 0 is enforced, the negative-reflection center.
! The reason only lx1 layers need to be negatively reflected is: this suffices for 
! the inner pixel at x1=x0+1.  Consider it's left arm is of length lx, that extends 
! beyond x0 by lx1.
! Typically,   x0 = x1-1, x7 = x6+1, etc.
real,      intent(inout) :: u(x0-lx1: x7+lx1, z0-lx1: z7+lx1)

!$OMP parallel do private(k)
do i = z0, z7
  u(x0, i)= 0.
  u(x7, i)= 0.
  !DEC$ IVDEP
  do k = 1, lx1
    u(x0-k, i) = -u(x0+k, i)
    u(x7+k, i) = -u(x7-k, i)
  end do
end do
!$OMP end parallel do

!$OMP parallel do private(k)
do i = x0, x7
  u(i, z0) = 0.
  u(i, z7) = 0.
  !DEC$ IVDEP
  do k = 1, lx1
    u(i, z0-k) = -u(i, z0+k)
    u(i, z7+k) = -u(i, z7-k)
  end do
end do
!$OMP end parallel do
end subroutine
!-------------
subroutine load_FDTD_boundary(u,  it)
integer,   intent(in)    ::       it
real,      intent(inout) ::   u(xb1: xb6, zb1: zb6)  ! recons. wavefield
!$OMP parallel do
do k = xb1, xb6
  u(k, zb1: zb2) = Bnd%u(k, zb1: zb2, it)
  u(k, zb5: zb6) = Bnd%d(k, zb5: zb6, it)
end do
!$OMP end parallel do

!$OMP parallel do
do k = zb3, zb4
  u(xb1: xb2, k) = Bnd%l(xb1: xb2, k, it)
  u(xb5: xb6, k) = Bnd%r(xb5: xb6, k, it)
end do
!$OMP end parallel do
end subroutine
!-------------
subroutine save_FDTD_boundary(u,  it)
integer,   intent(in)    ::       it
real,      intent(inout) ::   u(x1-lx: x6+lx, z1-lx: z6+lx) ! forwd. field
!$OMP parallel do
do k = xb1, xb6
  Bnd%u(k, zb1: zb2, it) = u(k, zb1: zb2)
  Bnd%d(k, zb5: zb6, it) = u(k, zb5: zb6)
end do
!$OMP end parallel do

!$OMP parallel do
do k = zb3, zb4
  Bnd%l(xb1: xb2, k, it) = u(xb1: xb2, k)
  Bnd%r(xb5: xb6, k, it) = u(xb5: xb6, k)
end do
!$OMP end parallel do
end subroutine
!-------------
subroutine wave_prop_1step(u, v, phi,  eta)
! run one step of FDTD, on wavefields {u, v} --> u
! phi  & eta   for PML
! lap, hdx_2, k, i      global variables
real,   intent(inout):: u(x1-lx: x6+lx, z1-lx: z6+lx), phi(x1:x6, z1:z6)
real,   intent(in)   :: v(x1-lx: x6+lx, z1-lx: z6+lx), &
                      eta(x1-1:  x6+1,  z1-1:  z6+1)
if (lx==2) then ! FDTD O(2,4)
!$OMP PARALLEL DO PRIVATE(I, LAP)
	do k=z1,z6
	   if ((k>=z3).and.(k<=z4)) then
		  !DEC$ IVDEP
		  do i=x1,x2
			 !! DAMPING LEFT.
			 lap=  coef(0)* v(i,  k)                   &
				  +coef(1)*(v(i+1,k)+v(i-1,k)          &
						   +v(i,k+1)+v(i,k-1))         &
				  +coef(2)*(v(i+2,k)+v(i-2,k)          &
						   +v(i,k+2)+v(i,k-2))

			 !! Update the wavefield.
			 u(i,k)=((2.-eta(i,k)*eta(i,k)             &
				  +2.*eta(i,k))*v(i,k)                 &
				  -u(i,k)                              &
				  +vdt2(i,k)*(lap                      &
				  +phi(i,k)))/(1.+2.*eta(i,k))

			 !! Update the PML function.
			 phi(i,k)=(phi(i,k)-                       &
				  ((eta(i+1,k)-eta(i-1,k))             &
				  *(v(i+1,k)-v(i-1,k))*hdx_2           &
				  +(eta(i,k+1)-eta(i,k-1))             &
				  *(v(i,k+1)-v(i,k-1))*hdx_2           &
				  ))/(1.+eta(i,k))
		  enddo
		  !DEC$ IVDEP
		  do i=x3,x4
			 !! BASIC.
			 lap=  coef(0)* v(i,  k)                   &
				  +coef(1)*(v(i+1,k)+v(i-1,k)          &
						   +v(i,k+1)+v(i,k-1))         &
				  +coef(2)*(v(i+2,k)+v(i-2,k)          &  
						   +v(i,k+2)+v(i,k-2))           

			 !! Update the wavefield.         
			 u(i,k)=2.*v(i,k)-u(i,k)+vdt2(i,k)*lap
		  enddo
		  !DEC$ IVDEP
		  do i=x5,x6
			 !! DAMPING RIGHT.
			 lap=  coef(0)* v(i,k)                     &
				  +coef(1)*(v(i+1,k)+v(i-1,k)          &
						   +v(i,k+1)+v(i,k-1))         &
				  +coef(2)*(v(i+2,k)+v(i-2,k)          &
						   +v(i,k+2)+v(i,k-2))

			 !! Update the wavefield.
			 u(i,k)=((2.-eta(i,k)*eta(i,k)             &
				  +2.*eta(i,k))*v(i,k)                 &
				  -u(i,k)                              &
				  +vdt2(i,k)*(lap                      &
				  +phi(i,k))) / (1.+2.*eta(i,k))

			 !! Update the PML function.
			 phi(i,k)=(phi(i,k)-                       &
				  ((eta(i+1,k)-eta(i-1,k))             &
				  *(v(i+1,k)-v(i-1,k))*hdx_2           &
				  +(eta(i,k+1)-eta(i,k-1))             &
				  *(v(i,k+1)-v(i,k-1))*hdx_2           &
				  ))/(1.+eta(i,k))
		  enddo
	   else
		  !DEC$ IVDEP
		  do i=x1,x6
			 !! DAMPING TOP and BOTTOM.
			 lap=  coef(0)* v(i,k)                     &
				  +coef(1)*(v(i+1,k)+v(i-1,k)          &
						   +v(i,k+1)+v(i,k-1))         &
				  +coef(2)*(v(i+2,k)+v(i-2,k)          &
						   +v(i,k+2)+v(i,k-2))

			 !! Update the wavefield.
			 u(i,k)=((2.-eta(i,k)*eta(i,k)             &
				  +2.*eta(i,k))*v(i,k)                 &
				  -u(i,k)                              &
				  +vdt2(i,k)*(lap                      &
				  +phi(i,k)))/(1.+2.*eta(i,k))

			 !! Update the PML function.
			 phi(i,k)=(phi(i,k)-                       &
				  ((eta(i+1,k)-eta(i-1,k))             &
				  *(v(i+1,k)-v(i-1,k))*hdx_2           &
				  +(eta(i,k+1)-eta(i,k-1))             &
				  *(v(i,k+1)-v(i,k-1))*hdx_2           &
				  ))/(1.+eta(i,k))
		  enddo
	   endif
	enddo ! k=z1,z6     --------- 
!$OMP end parallel do
else       ! FDTD O(2,8)
!$OMP PARALLEL DO PRIVATE(I, LAP)   
  do k=z1,z6
	if ((k>=z3).and.(k<=z4)) then
	   !DEC$ IVDEP
	   do i=x1,x2
		  !! DAMPING LEFT.
		  lap=  coef(0)* v(i,  k)                      &
			   +coef(1)*(v(i+1,k)+v(i-1,k)             &
						+v(i,k+1)+v(i,k-1))            &
			   +coef(2)*(v(i+2,k)+v(i-2,k)             &
						+v(i,k+2)+v(i,k-2))            &
			   +coef(3)*(v(i+3,k)+v(i-3,k)             &
						+v(i,k+3)+v(i,k-3))            &
			   +coef(4)*(v(i+4,k)+v(i-4,k)             &
						+v(i,k+4)+v(i,k-4))

		  !! Update the wavefield.
		  u(i,k)=((2.-eta(i,k)*eta(i,k)                &
			   +2.*eta(i,k))*v(i,k)                    &
			   -u(i,k)                                 &
			   +vdt2(i,k)*(lap                         &
			   +phi(i,k)))/(1.+2.*eta(i,k))

		  !! Update the PML function.
		  phi(i,k)=(phi(i,k)-                          &
			   ((eta(i+1,k)-eta(i-1,k))                &
			   *(v(i+1,k)-v(i-1,k))*hdx_2              &
			   +(eta(i,k+1)-eta(i,k-1))                &
			   *(v(i,k+1)-v(i,k-1))*hdx_2              &
			   ))/(1.+eta(i,k))
	   enddo
	   !DEC$ IVDEP
	   do i=x3,x4
		  !! BASIC.
		  lap=  coef(0)* v(i,  k)                      &
			   +coef(1)*(v(i+1,k)+v(i-1,k)             &
						+v(i,k+1)+v(i,k-1))            &
			   +coef(2)*(v(i+2,k)+v(i-2,k)             &
						+v(i,k+2)+v(i,k-2))            &
			   +coef(3)*(v(i+3,k)+v(i-3,k)             &
						+v(i,k+3)+v(i,k-3))            &
			   +coef(4)*(v(i+4,k)+v(i-4,k)             &
						+v(i,k+4)+v(i,k-4))

		  !! Update the wavefield.         
		  u(i,k) =2.*v(i,k)-u(i,k)+vdt2(i,k)*lap
	   enddo
	   !DEC$ IVDEP
	   do i=x5,x6
		  !! DAMPING RIGHT.
		  lap=  coef(0)* v(i,  k)                      &
			   +coef(1)*(v(i+1,k)+v(i-1,k)             &
						+v(i,k+1)+v(i,k-1))            &
			   +coef(2)*(v(i+2,k)+v(i-2,k)             &
						+v(i,k+2)+v(i,k-2))            &
			   +coef(3)*(v(i+3,k)+v(i-3,k)             &
						+v(i,k+3)+v(i,k-3))            &
			   +coef(4)*(v(i+4,k)+v(i-4,k)             &
						+v(i,k+4)+v(i,k-4))

		  !! Update the wavefield.
		  u(i,k)=((2.-eta(i,k)*eta(i,k)                &
			   +2.*eta(i,k))*v(i,k)                    &
			   -u(i,k)                                 &
			   +vdt2(i,k)*(lap                         &
			   +phi(i,k)))/(1.+2.*eta(i,k))

		  !! Update the PML function.         
		  phi(i,k)=(phi(i,k)-                          &
			   ((eta(i+1,k)-eta(i-1,k))                &
			   *(v(i+1,k)-v(i-1,k))*hdx_2              &
			   +(eta(i,k+1)-eta(i,k-1))                &
			   *(v(i,k+1)-v(i,k-1))*hdx_2              &
			   ))/(1.+eta(i,k))
	   enddo
	else
	   !DEC$ IVDEP
	   do i=x1,x6
		  !! DAMPING TOP and BOTTOM.
		  lap=  coef(0)* v(i,  k)                      &
			   +coef(1)*(v(i+1,k)+v(i-1,k)             &
						+v(i,k+1)+v(i,k-1))            &
			   +coef(2)*(v(i+2,k)+v(i-2,k)             &
						+v(i,k+2)+v(i,k-2))            &
			   +coef(3)*(v(i+3,k)+v(i-3,k)             &
						+v(i,k+3)+v(i,k-3))            &
			   +coef(4)*(v(i+4,k)+v(i-4,k)             &
						+v(i,k+4)+v(i,k-4))

		  !! Update the wavefield.
		  u(i,k)=((2.-eta(i,k)*eta(i,k)                &
			   +2.*eta(i,k))*v(i,k)                    &
			   -u(i,k)                                 &
			   +vdt2(i,k)*(lap                         &
			   +phi(i,k)))/(1.+2.*eta(i,k))

		  !! Update the PML function.         
		  phi(i,k)=(phi(i,k)-                          &
			   ((eta(i+1,k)-eta(i-1,k))                &
			   *(v(i+1,k)-v(i-1,k))*hdx_2              &
			   +(eta(i,k+1)-eta(i,k-1))                &
			   *(v(i,k+1)-v(i,k-1))*hdx_2              &
			   ))/(1.+eta(i,k))
	   enddo
	endif  ! if k ...
  enddo  ! k=z1,z6     --------- 
!$OMP end parallel do  
endif  ! FDTD order selection
end subroutine
!-------------
subroutine Laplace_nx_nz(v)
! compute  the Laplacian on v, in the physical domain nx x nz
! inner region using higher order, peripheral O(2)
! output:  Laplac(:,:),  the global var
real,      intent(inout) :: v(xb1: xb6, zb1: zb6)

if (lx==2) then  !!!!!!!!!!!!! O(2,4)
!$OMP PARALLEL DO PRIVATE(I)
  do k = z3, z4
	if(k>=zb3 .and. k<=zb4) then  ! central row, 3 cases:
	   !DEC$ IVDEP
	   do i = x3, xb2  !! left edge
		  Laplac(i,k)= (v(i+1,k)+v(i-1,k)+v(i,k+1)+v(i,k-1)-4*v(i,k))*dx_2
	   end do
	   !DEC$ IVDEP
	   do i= xb3, xb4  !! middle region
		  Laplac(i,k)=  coef(0) *v(i,  k)         &
			  +coef(1)*(v(i+1,k)+v(i-1,k)         &
					   +v(i,k+1)+v(i,k-1))        &
			  +coef(2)*(v(i+2,k)+v(i-2,k)         &  
					   +v(i,k+2)+v(i,k-2))           
	   end do
	   !DEC$ IVDEP
	   do i= xb5, x4   !! right edge
		  Laplac(i,k)= (v(i+1,k)+v(i-1,k)+v(i,k+1)+v(i,k-1)-4*v(i,k))*dx_2
	   end do
	else
	   !DEC$ IVDEP
	   do i=x3, x4
		  Laplac(i,k)= (v(i+1,k)+v(i-1,k)+v(i,k+1)+v(i,k-1)-4*v(i,k))*dx_2
	   end do
	end if  ! k \in [.,.]
  end do  ! k = ...
!$OMP end parallel do

else  !!!!!!!!!!!!! O(2,8)
!$OMP PARALLEL DO PRIVATE(I)
do k = z3, z4
	if(k>=zb3 .and. k<=zb4) then  ! central row, 3 cases:
	   !DEC$ IVDEP
	   do i = x3, xb2  !! left edge
		  Laplac(i,k)= (v(i+1,k)+v(i-1,k)+v(i,k+1)+v(i,k-1)-4*v(i,k))*dx_2
	   end do
	   !DEC$ IVDEP
	   do i= xb3, xb4  !! middle region
		  Laplac(i,k)=   coef(0) *v(i,  k)             &
			   +coef(1)*(v(i+1,k)+v(i-1,k)             &
						+v(i,k+1)+v(i,k-1))            &
			   +coef(2)*(v(i+2,k)+v(i-2,k)             &
						+v(i,k+2)+v(i,k-2))            &
			   +coef(3)*(v(i+3,k)+v(i-3,k)             &
						+v(i,k+3)+v(i,k-3))            &
			   +coef(4)*(v(i+4,k)+v(i-4,k)             &
						+v(i,k+4)+v(i,k-4))         
	   end do
	   !DEC$ IVDEP
	   do i= xb5, x4   !! right edge
		  Laplac(i,k)= (v(i+1,k)+v(i-1,k)+v(i,k+1)+v(i,k-1)-4*v(i,k))*dx_2
	   end do
	else
	   !DEC$ IVDEP
	   do i=x3, x4
		  Laplac(i,k)= (v(i+1,k)+v(i-1,k)+v(i,k+1)+v(i,k-1)-4*v(i,k))*dx_2
	   end do
	end if  ! k \in [.,.]
  end do  ! k = ...
!$OMP end parallel do
end if
end subroutine
!-------------
subroutine wave_1step_backprop(u, v, bn, e, it) ! having computed the Laplacian
! Fulcrum V points to slice it.  U will be updated to it-1, from it+1
! for reconstructing source wavefield.

real,      intent(inout)    :: u(xb1: xb6, zb1: zb6)
real,      intent(inout)    :: v(xb1: xb6, zb1: zb6)
type(bund_type),   intent(in)   ::   bn
type(entire_param),intent(in)   ::    e
integer,           intent(in)   ::   it

!$OMP PARALLEL DO PRIVATE(I)
 do k = z3, z4
   !DEC$ IVDEP
   do i = x3, x4
      u(i,k) = 2.*v(i,k) - u(i,k) + vdt2(i,k) * Laplac(i,k)
   end do
 end do
!$OMP end parallel do

! remove sources originally injected on v:
call inject_sources(  v, xb1,xb6, zb1,zb6, it, bn, e, vdt2, .true.)
call load_FDTD_boundary(u, it-1)
end subroutine
!-------------
subroutine grd_accumu(grd, de)
real,    intent(inout):: grd( x3  :x4   ,     z3 :z4   )
real,    intent(in)   :: de(x1-lx :x6+lx,  z1-lx :z6+lx) ! back-proj data err
!$OMP PARALLEL DO PRIVATE(I)
 do k = z3, z4
   !DEC$ IVDEP
   do i = x3, x4
      grd(i,k) = grd(i,k) + de(i,k) * Laplac(i,k)
   end do
 end do
!$OMP end parallel do
end subroutine
!-------------
subroutine Precnd_accumu(Grn2)
real,    intent(inout):: Grn2(x3:x4,z3:z4)
!$OMP PARALLEL DO PRIVATE(I)
 do k = z3, z4
   !DEC$ IVDEP
   do i = x3, x4
      Grn2(i,k) = Grn2(i,k) + Laplac(i,k) * Laplac(i,k)
   end do
 end do
!$OMP end parallel do
end subroutine
!################################

subroutine forward_a_bundle(csgc, nt, bn, e)
real,              intent(out)  :: csgc(:,:)
integer,           intent(in)   :: nt ! simu. steps (may /= md%nt)
type(bund_type),   intent(in)   :: bn
type(entire_param),intent(in)   :: e
! Delay possible preconditioner computation (\Sum |wavefield|^2) to the gradient
! routine, where inner domain of source wavefield is being reconstructed, whose
! size is the same as that of the preconditioner.
logical(1),save    :: forw_1stcall = .true.
integer            :: tt, t2

wvext1=0.    ;     wvext2=0.    ;     phi  = 0.
p2  => wvext1;     p1  => wvext2
!          WORKFLOW:
!   @ t
! 1. ... propagate to here 
!        [a) p2=>here; b) swap(p2,p1); so, p1=>here, p2=>last slice]
! 2. inject  sources
! 3. harvest receivers
! 4. impose  free surface condition, to prepare for propagation to next step ...
! 5. save    boundaries (in reconstruction, this place is what's reconstructed)
!---------- 
!   Initially,  p2=>0,    p1=>[t=1]|_0  this state is at the end of (1).
!   At last,    p2=>nt-1, p1=>nt, w/ inj. & harv.; is at the end of (4).
!---------- Re-wrap the steps:
! 0: init.
! [2 -> 3 -> 4 -> 5'-> 1],  5': exit if t==nt
tt  = 1
do
	! step 2:
	call inject_sources(p1, x1-lx,x6+lx,z1-lx,z6+lx, tt, bn, e, vdt2)
	! step 3:
	call harvest_rcvrs( csgc, p1, tt, bn, e)
	! step 4:
	call put_free_surface(p1, lx-1, x1-1, x6+1, z1-1, z6+1)
	! step 5:
	t2 = tt;  if(e%o%MUTE==2) t2 = tt - e%d%nt   ! only record the 2nd half
	if(t2>0 .and.t2<e%d%nt-1) call save_FDTD_boundary(p1, t2)
#ifdef DEBUGfwi
 if(forw_1stcall.and.rank==0) call write_binfile('snap/forw0.bin',p1(x3:x4,z3:z4),-1)
#endif
    if (tt >= nt)  exit
    ! step 1:
    call wave_prop_1step(p2,p1, phi,eta)
	ptmp => p2;    p2 => p1;    p1 => ptmp	
	tt   =  tt + 1
end do
if(forw_1stcall) forw_1stcall = .false.
end subroutine
!################################
SUBROUTINE INIT_GRADIENT(RTMIMG, op, Green2) ! call before accruing the grad
real,               intent(out)  :: RTMIMG(:,:)
type(option_param), intent(in)   :: op
real,optional,      intent(out)  :: Green2(:,:)
RTMIMG = 0.
if (op%PRECND==2) then
 if (.not. present(Green2))   stop 'init_gradient() expects Green2!'
 Green2 = 0. 
end if
END SUBROUTINE
!-------------
subroutine gradient_a_bundle(img, gtmptr, nt, csge, bn, e, Grn2)
! to produce N E G A T I V E  gradient (to be accumulated) wrt to slowness,
!    missing a 2/s factor.
! gtmptr points to temp storage, for MPIreduction: grd and Grn2
! input:  csge \def= calc - obsrv ~ L m - d
real,              intent(inout)  ::   img(:,:)
real,              pointer        ::   gtmptr(:,:)
integer,           intent(in)     ::   nt
real,              intent(in)     ::   csge(:,:)
type(bund_type),   intent(in)     ::   bn
type(entire_param),intent(in)     ::   e
real,optional,     intent(inout)  ::   Grn2(:,:)

logical(1),save   :: back_1stcall = .true.
integer           :: tt

! A) init.
! 1. saving forwd. modeling source wavefield's final 2 slices, w/o pml:
wvfld1 = p2(xb1 :xb6, zb1 :zb6);   pb   => wvfld1 ! nt-1, fulcrum
wvfld2 = p1(xb1 :xb6, zb1 :zb6);   pa   => wvfld2 ! nt,   prop update. then pb--> working slice
! deducting source from nt:
call inject_sources( pa, xb1,xb6, zb1,zb6, nt, bn, e, vdt2, .true.)

! 2. init. data error back-prop
wvext1=0.    ;     wvext2=0.    ;     phi  = 0.
p2  => wvext1;     p1  => wvext2
call inject_misfit(p1, nt, bn, csge, e, vdt2)   ! p1: fulcrum;  p2: prop update
call put_free_surface(p1, lx-1, x1-1, x6+1, z1-1, z6+1)  ! then p1--> the working slice

if(e%o%MUTE==2) then    ! back-prop by e%d%nt
  do tt = nt-1, e%d%nt, -1
    call backprop1step_del_d(tt)					  
  end do    ! now, p1 points to e%d%nt, w/ injection;  p2 to e%d%nt+1
end if

! B) main do, in lock step
tt = e%d%nt-1
do 
    call Laplace_nx_nz(pb)   ! pb--> tt;   pa--> tt+1
	call grd_accumu(img, p1) ! p1--> tt+1; p2--> tt+2
	if (e%o%PRECND==2) call Precnd_accumu(Grn2)
	if (tt==1)         exit
    call backprop1step_del_d(tt)               ! p1 --> tt;   p2--> tt+1
	call wave_1step_backprop(pa, pb, bn, e, tt)! pb --> tt;   pa--> tt-1
	ptmp => pa;     pa => pb;     pb => ptmp   ! pb --> tt-1; pa--> tt
	tt = tt - 1
#ifdef DEBUGfwi
  if(back_1stcall .and. rank==0) then ! 
	call write_binfile('snap/recons0.bin', pb(x3:x4,z3:z4), -1) ! nt-2 --> 1
	call write_binfile('snap/back0.bin',   p2(x3:x4,z3:z4), -1) ! nt --> 3
  end if   
#endif    
end do

gtmptr => Laplac
#ifdef DEBUGfwi
  if (back_1stcall) then
    call filename(tmpfnm,       './snap/grd', rank, '.bin' )
    call write_binfile(tmpfnm,   img ) 
  end if
#endif
if (back_1stcall)   back_1stcall = .false.

contains
  subroutine  backprop1step_del_d(it)
	integer,  intent(in) :: it
    ! WORKFLOW:
	! 1. prop to time step 'it'. 
	!    i) p1 being the fulcrum at it+1
    !   ii)	p2 is updated from it+2 to it, using the wave equation.
    !  iii) swap(p1, p2).  So, p1 points to it;  p2 to it+1
	! 2. inject data error, for the working slice P1, i.e., at IT
	! 3. impose free surface b.c.
	call wave_prop_1step( p2, p1, phi,eta)                   ! step 1
	ptmp => p2;     p2 => p1;     p1 => ptmp
    call inject_misfit(   p1, it, bn, csge, e, vdt2)         ! step 2
	call put_free_surface(p1, lx-1, x1-1, x6+1, z1-1, z6+1)  ! step 3
  end subroutine
end subroutine
!===================================
END MODULE
