! Copyright (C) 2012 Center for Subsurface Imaging and Fluid Modeling (CSIM),
! King Abdullah University of Science and Technology, All rights reserved.
! 
! Sponsors of CSIM are granted a non-exclusive, irrevocable royalty free
! world-wide license to use this software and associated documentation files
! (the "Software"), in their business, including the rights to modify and
! distribute the Software to their affiliates, partners, clients or consultants
! as necessary in the conduct of the sponsors business, all without accounting
! to the King Abdullah University of Science and Technology, subject to the
! following conditions:
! 
! The above copyright notice and this permission notice shall be included
! in all copies or substantial portions of the Software.
! 
! Warranty Disclaimer:
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
! OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
! FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
! THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
! LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
! FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
! DEALINGS IN THE SOFTWARE.
!-----------------------
!  version:     1.0
!  author:      Yunsong Huang
!  email:       yunsong.huang@kaust.edu.sa
!  updated:     01-29-2013
!  purpose:     {FWI, forward modeling} x {marine, fixed-spread}_geometry, using 2d FDTD with PML
!               Notation: {...} denotes a set;  'x' denotes direct product
!-----------------------


! slw:  gradient (and update) w.r.t. slowness  (grd is -Grad)
! mar:  multiSource marine towed streamer
! x:    fast dim.

program fwi2d_mar_slw
use nrtype;             use myPARSER
use myMath_module;      use seism_uti_module
use myOptim_mod;        use myString_mod
use myPML;              use myMPI0
use datatype0;          use myIO
use source_multi;       use fdtd_2d
use datatype1
#ifdef _OPENMP
 use OMP_LIB
#endif

implicit none

type(entire_param)      :: e
type(proc_type)         :: pr

real,target,allocatable :: g_mem(:,:,:), q_mem(:,:,:)
                         !  a x b x 2,  memory space, use pointer swapping:
real,pointer            :: pg_old(:,:), pg_cur(:,:),  pq_ext(:,:), pq_nxt(:,:),&
                           pd(:,:)    ! preconditioned -grd_current
						 !  g: -grad,   old & current
						 !  q: C.G. search dir, existing & next	 
real,allocatable        :: v_tru(:,:),        &
                           prcnd(:,:),        &  ! 1/[\Sum_{t,s} G(r,t|s)^2] 
                           slow_old(:,:),     &  ! existent v's reciprocal
                           v_try(:,:)            ! current  trial v                      
integer                 :: it, 		maxIt, 	existingIt
real                    :: t1, 		t2
double precision        :: ompt1,   ompt2
character(len=100)      :: tmpfnm
! NON-ROOT only needs:  v_try, slow_old, pq_ext

!integer   :: DebugWait
!DebugWait = 1
!do while(DebugWait)
!  it = 0  ! dummy
!end do


call   init_all

if (   e%o%FWI /= 0)    then
  call main_update
else
  call forward(    v_try     )  
end if

#ifdef DEBUGfwi
 call  probe_grad
 call  stop_mpi; stop
#endif

call   kill_all
contains

!=========================================
subroutine         kill_all
 call    destruct_types(  e, pr )
 deallocate(g_mem, q_mem, v_tru, prcnd, slow_old, v_try)
 call    kill_pml
 call    kill_fdtd
 call    stop_mpi
end subroutine

subroutine         INIT_ALL
 integer                 :: omp_id, omp_n_thre
 real                    :: memsz, szcsg, szfwi, szpml, szfdtd
 character(len=100)      :: parfile
 
 call   init_mpi
 if (rank==0)  call cpu_time(t1)

 call   readCmd('par', parfile, 'parfile')
 call   read_parfile(  parfile, maxIt,  existingIt)
 
 call   construct_types(e, pr, parfile, existingIt, 'log/fwi_mar.log', szcsg)
 
 call   alloc_mem(szfwi)
 
 if (e%i%V_TRU_AVAIL)  call IO_model(v_tru, e%i%vel_tru_file, e%i%XFAST, .true.)

 call init_pml( e%d%vmax,e%d%dx,e%d%dt,e%d%nx,e%d%nz,e%d%npml,e%o%FREESURF==1,szpml)
 szfdtd = init_fdtd(  e%d  )
 call    IO_model(v_try, e%i%vel_ini_file,    e%i%XFAST, .true.)
 call    bcast(v_try)
 slow_old = 1./v_try
 call update_vdt2(v_try, e%d%dt, e%d%npml, e%d%nx, e%d%nz, e%o%FREESURF)
  
 memsz    = szcsg + szfwi + szpml + szfdtd
 write(*,*) 'Proc id, memory(MB) for [csg, fwi_optim, pml, fdtd] & total:', rank, &
  (/ szcsg, szfwi, szpml, szfdtd /)*1e-6, memsz*1e-6
 call flush(6)
 if (rank==0)  then
   call cpu_time(t2); write(*,*) 'preproc. time: ', t2-t1, '(s)'
   call flush(6)
   t1 = t2
 end if
!$omp parallel PRIVATE(omp_id)
!$  omp_id     = omp_get_thread_num()
!$  omp_n_thre = omp_get_num_threads()
!$  write(*,*)  'I am process ', rank, ', thread ', omp_id, &
!$              ', among ', omp_n_thre, ' threads in total.'
!$  if (rank==0) ompt1 = omp_get_wtime()
!$omp end parallel 
 
#ifdef DEBUGfwi
   if (rank==0) then
     call write_binfile('snap/v_try.dat', v_try)
     call write_binfile('snap/eta.dat', eta)
   end if  
#endif 
end subroutine


subroutine  alloc_mem(memsz)
real,       intent(out)    :: memsz
if (rank==0) then
  allocate( g_mem(e%d%x3:e%d%x4, e%d%z3:e%d%z4,3), &
            q_mem(e%d%x3:e%d%x4, e%d%z3:e%d%z4,2))
  pg_old => g_mem(:,:,1);  pg_cur => g_mem(:,:,2);   pd => g_mem(:,:,3)
  pq_ext => q_mem(:,:,1);  pq_nxt => q_mem(:,:,2)
else
  allocate( g_mem(e%d%x3:e%d%x4, e%d%z3:e%d%z4,1), &
            q_mem(e%d%x3:e%d%x4, e%d%z3:e%d%z4,1))
  pg_old => g_mem(:,:,1);  pg_cur => g_mem(:,:,1);   pd => g_mem(:,:,1)
  pq_ext => q_mem(:,:,1);  pq_nxt => q_mem(:,:,1)
end if
if (e%o%PRECND==2) then
   allocate(prcnd(e%d%x3:e%d%x4, e%d%z3:e%d%z4) )
else
   allocate(prcnd(1,1))
end if

allocate( v_try(e%d%x3:e%d%x4, e%d%z3:e%d%z4), &
       slow_old(e%d%x3:e%d%x4, e%d%z3:e%d%z4) )

if ( e%i%V_TRU_AVAIL) then
  allocate( v_tru(e%d%x3:e%d%x4, e%d%z3:e%d%z4) )
else
  allocate( v_tru(0,0) ) ! dummy, to make it formally valid 
end if
memsz = ( size(g_mem) + size(q_mem) + size(prcnd) + size(v_try) + &
       size(slow_old) + size(v_tru) )*4.
end subroutine
!===============================================

subroutine perturb_v(v1, slw, d, COMMIT)   ! ALL PROC.'s DO
! slowness + d --> 1/(.) --> v1
! Note:   this could be a temporary move in the line search
! When COMMIT, slw is also updated
real,     intent(out)    ::   v1(e%d%x3:e%d%x4,e%d%z3:e%d%z4)
real,     intent(inout)  ::  slw(e%d%x3:e%d%x4,e%d%z3:e%d%z4)
real,     intent(in)     ::    d(e%d%x3:e%d%x4,e%d%z3:e%d%z4)
logical,  intent(in)     ::  COMMIT
integer,  save           :: it = 1
integer                  :: ix, iz
real                     :: smx, smn
smx = 1. / e%d%vmin
smn = 1. / e%d%vmax

! When COMMIT & @ every 40 iter., non-root nodes don't compute and instead wait
! for the Bcast from the root.  Otherwise, everyone computes.  Truth table:
! _________|__COMMIT&40__|_~(COMMIT&40)__
! rank==0  |    compu    |    compu
! rank/=0  |  no compu   |    compu
! ---------------------------------------
! Compu's condition is:  .not.(rank/=0 .and. COMMIT .and. MOD(it,40)==0)
! Bcast's condition is:  (COMMIT & 40th iter)

 if ( .not.(rank/=0 .and. COMMIT .and. MOD(it,40)==0) ) then
 !$OMP PARALLEL DO PRIVATE(ix)
   do iz   = e%d%z3, e%d%z4
     do ix = e%d%x3, e%d%x4
       v1(ix,iz)   = slw(ix, iz) + d(ix, iz) ! v1 is slowness here
       if(v1(ix,iz)> smx)     then
         v1(ix,iz) = smx
       else if(v1(ix,iz)<smn) then
         v1(ix,iz) = smn
       end if
       v1(ix,iz) = 1./v1(ix,iz)  ! now is v
     end do
   end do
 !$OMP end parallel do  
 end if
 if (COMMIT .and. MOD(it,40)==0)   call bcast(v1)
 call update_vdt2(v1, e%d%dt, e%d%npml, e%d%nx, e%d%nz, e%o%FREESURF)
 if (COMMIT)  slw = 1./v1
 it   =  it + 1
end subroutine

subroutine precond_grad(dir, grd, precn, v)               ! ONLY ROOT CALLs
! grd & dir are wrt slowness. 
! inputs  pertain to the central region, while
real,     intent(out)  :: dir(e%d%x3:e%d%x4,e%d%z3:e%d%z4)  ! precon search dir
real,     intent(in)   :: grd(e%d%x3:e%d%x4,e%d%z3:e%d%z4), &
                        precn(e%d%x3:e%d%x4,e%d%z3:e%d%z4), &
                            v(e%d%x3:e%d%x4,e%d%z3:e%d%z4)
integer   :: ix, iz
!! the preconditioner affects convergence speed crucially!!
!! As my precn uses |Laplac{wavefield}|^-2,  it effectively has a v^4,
!! which is too much.  So divide out v^2.  seems best.
!$OMP PARALLEL DO PRIVATE(ix)
do iz   = e%d%z3, e%d%z4
  do ix = e%d%x3, e%d%x4
    if (e%o%PRECND==2) then
      dir(ix, iz) = precn(ix, iz) * grd(ix, iz) ! Green Precond.
    else if (e%o%PRECND==1) then
      dir(ix, iz) = (iz - e%d%z3 + 1)**2 * grd(ix, iz)
    else  
      dir(ix, iz) = grd(ix, iz)
    end if 
    dir(ix, iz) = dir(ix, iz) / v(ix, iz)**2
  end do
end do
!$OMP end parallel do 
end subroutine

function  step_length(stp_lst, it, dir)  result(stpsz) ! ONLY ROOT CALLs
real                   :: stpsz
integer,  intent(in)   :: it                ! current iteration
real,     intent(in)   :: stp_lst, dir(:,:) ! last optim stp len, current search dir

real,     save         :: stp_pprv   ! the previous (before last) optim stp len
integer,  save         :: Ncalled=0  ! # of times this funct. has been called
integer,  parameter    :: it00=1

if (Ncalled== 0) then
  stpsz    = (1+it00)/real(it+it00) *(0.02/e%d%vmin) / (norm_1(dir)/e%d%nz/e%d%nx)
  stp_pprv = stpsz
else
  stpsz    = sqrt(stp_lst * stp_pprv) ! geometrical avrg of previous two steps
  stp_pprv = stp_lst
end if
Ncalled    = Ncalled + 1
end function

!-------------
function misfit_driver(x)   result (f) ! f: dummy for non-root proc.'s
! driver for using 1d line search optimiz. routine
 real(SP),     intent(inout) :: x  ! only root has the valid x at input
 real(SP)                    :: f
 call MPI_BCAST(x, 1, MPI_REAL, 0, MPI_COMM_WORLD, ierr)
 call perturb_v(v_try, slow_old, x*pq_ext, .false.)    ! all proc.'s update
  ! call filename(tmpfnm, 'log/pv', rank)
  ! call write_binfile(tmpfnm, v_try)
 f  = forward_n_grd(v_try)                        ! all proc.'s participate
end function
!************************************************************
subroutine        main_update
 integer                  :: Nfeval
 real                     :: stpsz, misfit, misfit1, xmin=0.1, rec(20,2)
 logical                  :: hit_n_reset
 it = existingIt
 call randomize_encoder(pr, e)
 misfit = forward_n_grd(v_try, pg_cur) ! on newly encoded csg's
 call init_spit(misfit)

do it  = existingIt+1, maxIt
  if (rank==0)  call precond_grad(pd, pg_cur, prcnd, v_try)
  call  CG_precon_reset(pq_ext,pq_nxt, pg_old,pg_cur, pd, int(e%o%CGreset), &
        int(e%o%CGresetBlock), hit_n_reset=hit_n_reset, rk=rank)
  if (rank==0)  stpsz = step_length(xmin, it, pq_ext)
  
#ifdef DEBUGfwi
  misfit1= optimz_1d(xmin, 0., stpsz,misfit, misfit_driver, 1e-2, rank,Nfeval, rec)
  if (rank==0 .and. it==maxIt) call write_tex('./log/lineSearch.log',rec(1:Nfeval,:))
#else
  misfit1= optimz_1d(xmin, 0., stpsz,misfit, misfit_driver, 1e-2, rank,Nfeval)
#endif

  call MPI_BCAST(xmin, 1, MPI_REAL, 0, MPI_COMM_WORLD, ierr)
  call perturb_v(v_try,  slow_old,  xmin*pq_ext,     .true.)
  call randomize_encoder( pr, e)
  if (it/=maxIt) then
    misfit = forward_n_grd(v_try, pg_cur) 
  else
    misfit = forward_n_grd(v_try        ) 
  end if
  call spit_out(misfit)
enddo
end subroutine
!*************
	subroutine init_spit(misfit)
	real, intent(in) :: misfit
	real             :: verr
		if (rank/=0)       return
		
#ifdef DEBUGfwi
		call showArryInfo(prcnd, 'prcnd')
		call write_binfile('log/precnd.bin', prcnd)
#endif
		if (existingIt/=0) return
		
		call save_stuff(verr,misfit, v_try, it,e%i,e%o, v_tru, pg_cur)
	end subroutine
	
	subroutine spit_out(misfit)
	real, intent(in) :: misfit
	real             :: verr
	    if (rank/=0)       return
		call save_stuff(verr,misfit, v_try, it,e%i,e%o, v_tru, pg_cur)
		call cpu_time(t2)
		
		write(*,263) (t2-t1)/60, (t2-t1)*(maxIt-it)/(it-existingIt)/60.
        263 format('Time  spent  &  left   ', 2ES12.3, '(min) ')
	
		if (it==maxIt)      write(*,*) 'Runtime / Iter = ', &
		             (t2-t1)/60./(it-existingIt),      '(min)'
					 
		write(*,*) '-----------------------';   call flush(6)
	end subroutine
!************************************************************

function forward_n_grd(v, grd)  result(fval) ! grd & fval: only valid @ root
! fval:         misfit  || d_cal - d_obs ||^2
real,           intent(in)   ::  v(:,:)
real, optional, pointer      ::  grd(:,:)  ! if absent, then no precond
! grd must have space allocated, for every processs
real,           pointer      ::  gtmp(:,:)
real                         ::  fval,  fval_loc, ftmp
integer                      ::  ib, nt

 nt = 2* e%d%nt                 ! if marine towed multisource --> freq. selec.
 if(e%o%MUTE /= 2)  nt = e%d%nt ! else, fix-spread or single source
 
 if (present(grd))  call INIT_GRADIENT(grd, e%o, prcnd)
 fval_loc = 0.;     fval = 0.
 do ib = pr%bi1,    pr%bi2
  if (e%o%SINU==0)  call load_n_encode_csg(ib, pr, e)
  call forward_a_bundle(pr%csg%c, nt, pr%b(ib), e)
  ftmp = compute_misfit(pr%csg,       pr%b(ib), e)
  fval_loc= fval_loc+ ftmp
  if (present(grd)) call gradient_a_bundle(grd,gtmp,nt, pr%csg%e, pr%b(ib),e,prcnd)
 end do
 call MPI_REDUCE(fval_loc,fval,1, MPI_REAL, MPI_SUM, 0,MPI_COMM_WORLD,ierr)
if (.not. present(grd))  return
 
if(e%o%PRECND==2) then  ! note: PRCND is available ONLY @ ROOT
  call MPI_REDUCE(prcnd,gtmp,e%d%nz*e%d%nx,MPI_REAL,MPI_SUM,0,MPI_COMM_WORLD,ierr)
  if (rank==0)    prcnd = 1. / gtmp
end if

! GRD ~ search_dir in slowness,  available ONLY @ ROOT
call MPI_REDUCE(grd, gtmp, e%d%nz*e%d%nx,MPI_REAL,MPI_SUM,0,MPI_COMM_WORLD,ierr)

if (rank==0) then
  grd = gtmp *2.*v       ! make up the missing 2/s, see gradient_a_bundle()
  call freeze_water_velocity(grd, e%s%z_fix_v, e%s%waterBtm)
end if  
end function
!...........
subroutine forward(v)  ! generate csg, typically using true model. Single Source.
! use wavelet in t-domain.  If towed, then generated csg in freq. domain as well.
real,         intent(in) :: v(:,:)
integer       :: ib, nt, ng
if (rank==0)  then
  tmpfnm = trim(e%i%csgout_file) // '_SrcWvlt.bin'
  call write_binfile(tmpfnm,   e%s%s(::e%o%FWD_sbsmp) )  ! subsample
end if
 nt   = e%d%nt
 call   randomize_encoder(pr, e)
 do ib= pr%bi1, pr%bi2                     ! ib==is, 'coz single source
  call  forward_a_bundle(        pr%csg%c,   nt,   pr%b(ib), e)
  if(e%o%TOW==0)  then
    call filename(tmpfnm,        e%i%csgout_file, ib, '.bin' )
    call write_binfile(tmpfnm,   pr%csg%c(::e%o%FWD_sbsmp,:) ) 
  else
    ng = pr%b(ib)%rj2 - pr%b(ib)%rj1 + 1
    call filename(tmpfnm,        trim(e%i%csgout_file)//'M', ib, '.bin')  ! Marine, Ricker
    call write_binfile(tmpfnm,   pr%csg%c(::e%o%FWD_sbsmp,1:ng))
    
    call filename(tmpfnm,         trim(e%i%csgout_file)//'S', ib, '.bin') ! Sinu, freq.Divi
  	call convert2Fdomain(pr%csg%e,pr%csg%c, e%s%trctmp, nt, e%s%nw, ng)
  	call write_binfile(tmpfnm,    pr%csg%e(:,1:ng))
  end if
 end do
if (rank==0) then
!$omp parallel
!$  ompt2 = omp_get_wtime()
!$  write(*,*) 'OMP wall clock time: ', (ompt2 - ompt1)/60., '(min)'
!$omp end parallel 
  call cpu_time(t2)
  write(*,*) 'Time spent forward modeling: ', (t2-t1)/60., '(min)'
! cpu_time is the aggregate time of OMP processes, 
!  ~ omp_get_num_threads() * [diff_of omp_get_wtime()]
  call flush(6)
end if
end subroutine
!=====================
subroutine probe_grad
real    :: dels, misfit, misfit1, slp
integer :: zc, xc
real, pointer   :: gtmp(:,:) ! temp storage

if(rank==0) write(*,*) 'entering probe_grad'
call flush(6)
 call randomize_encoder(pr, e)
 call test__rand_enc(pr, e)

!if (rank == nsize-1) then
!	write(*,*) '-- testing iterator --'
!	write(*,*) 'bi1, bi2',  pr%bi1, pr%bi2
!	do zc = pr%bi1, pr%bi2
!	  call test__iter_src(pr%b(zc), e)
!	end do
!end if
zc   = 10
xc   = int(e%d%nx/2)
dels = 5e-3 * slow_old(xc, zc)

misfit = forward_n_grd(v_try, pg_cur) ! on newly encoded csg's
if(rank==0) write(*,*) 'misfit0: ',  misfit
call flush(6)

v_try(xc, zc) = 1./(slow_old(xc,zc) + dels)
call update_vdt2(v_try, e%d%dt, e%d%npml, e%d%nx, e%d%nz, e%o%FREESURF)
misfit1 = forward_n_grd(v_try)

if(rank==0) write(*,*) 'misfit1, slow_old(xc, zc), dels, v_try(xc,zc)', &
                        misfit1, slow_old(xc, zc), dels, v_try(xc,zc)
call flush(6)

v_try(xc, zc) = 1./(slow_old(xc,zc) - dels)
call update_vdt2(v_try, e%d%dt, e%d%npml, e%d%nx, e%d%nz, e%o%FREESURF)
misfit = forward_n_grd(v_try)

if(rank==0) write(*,*) 'misfit2, v_try(xc,zc)', misfit, v_try(xc,zc)
call flush(6)

if(rank==0) then
  slp = (misfit1 - misfit)/(4.*dels)
  write(*,*) 'misfit1, misfit, slp, grd, max|grd|', misfit1, misfit, slp, &
              -pg_cur(xc,zc), maxval(abs(pg_cur(:,3:)))
  call flush(6)
end if

end subroutine

end program
